# -*- coding:utf-8 -*-
from urllib import parse


def get_attributes_from_url(url: str) -> dict:
    """获取网址中的参数字典"""
    url_data = parse.unquote(url)  # url解码
    result = parse.urlparse(url_data)  # url结果
    return parse.parse_qs(result.query)  # url里的查询参数


def get_attribute_value_from_url(url: str, attribute: str) -> str:
    """获取网址中某参数的值"""
    results = get_attributes_from_url(url)
    if attribute not in results:
        return ''
    if len(results[attribute]) == 1:
        return results[attribute][0]
    else:
        return results[attribute]



# -*- coding:utf-8 -*-
"""
常用的excel文件处理方法
"""
# 安装win32com：pip install pypiwin32
from copy import copy
from pathlib import Path
from typing import Union

from openpyxl import Workbook
from openpyxl.cell import Cell
from openpyxl.utils import get_column_letter, column_index_from_string
from openpyxl.worksheet.worksheet import Worksheet

from .paths import make_valid_name, get_usable_path


def xls_to_xlsx(path: Union[str, Path], del_src: bool = True) -> Union[Path, bool]:
    """把xls格式另存为xlsx格式                  \n
    :param path: xls文件路径
    :param del_src: 另存之后是否删除源文件
    :return: 是否成功
    """
    from win32com import client
    excel = None
    try:
        excel = client.Dispatch("Ket.Application")  # 调用WPS程序
    except:
        try:
            excel = client.gencache.EnsureDispatch('Excel.Application')  # 调用Office程序
        except:
            pass

    if excel:
        path = str(Path(path).absolute())
        wb = excel.Workbooks.Open(path)
        new_name = get_usable_path(f"{path}x")
        wb.SaveAs(str(new_name), 51)
        wb.Close()
        if del_src:
            Path(path).unlink()
        return new_name

    return False


def copy_cell(cell_or_loc: Union[Cell, str],
              tar_sheet: Worksheet,
              row_or_loc: Union[str, int],
              col: Union[str, int] = None,
              src_sheet: Worksheet = None,
              copy_format: bool = True) -> None:
    """复制单元格到目标工作表                               \n
    :param cell_or_loc: 源单元格或其坐标（'B3'或'3,3'），输入坐标时必须输入src_sheet
    :param tar_sheet: 目标工作表
    :param row_or_loc: 目标工作表中单元格行号或位置字符串
    :param col: 目标工作表中单元格位置
    :param src_sheet: 源工作表
    :param copy_format: 是否复制格式
    :return: None
    """
    if isinstance(cell_or_loc, str):
        if not src_sheet:
            raise ValueError('用坐标定位时必须输入sheet')

        if ',' in cell_or_loc:
            loc = cell_or_loc.split(',')
            cell_or_loc = src_sheet.cell(int(loc[0].strip()), int(loc[1].strip()))
        else:
            cell_or_loc = src_sheet[cell_or_loc]

    if isinstance(row_or_loc, str):
        tar_cell = tar_sheet[row_or_loc]
    elif isinstance(row_or_loc, int) and isinstance(col, int):
        tar_cell = tar_sheet.cell(row_or_loc, col)
    elif isinstance(row_or_loc, int) and isinstance(col, str):
        tar_cell = tar_sheet[f'{col}{row_or_loc}']
    else:
        raise ValueError('传入的位置参数不正确')

    tar_cell.value = cell_or_loc.value
    tar_cell.hyperlink = cell_or_loc.hyperlink

    if copy_format and cell_or_loc.has_style:
        tar_cell.font = copy(cell_or_loc.font)
        tar_cell.border = copy(cell_or_loc.border)
        tar_cell.fill = copy(cell_or_loc.fill)
        tar_cell.number_format = copy(cell_or_loc.number_format)
        tar_cell.protection = copy(cell_or_loc.protection)
        tar_cell.alignment = copy(cell_or_loc.alignment)


def copy_row(row_or_num: Union[int, tuple],
             tar_sheet: Worksheet,
             src_sheet: Worksheet = None,
             tar_row_num: int = None,
             copy_col_width: bool = True) -> None:
    """复制一行到目标工作表                             \n
    :param row_or_num: 源行对象或行号
    :param tar_sheet: 目标工作表
    :param src_sheet: 源工作表
    :param tar_row_num: 目标行号，默认为在最后添加一行
    :param copy_col_width: 是否复制列宽
    :return: None
    """
    if not tar_row_num:
        max_row = tar_sheet.max_row
        if len(tar_sheet[max_row]) == 1 and tar_sheet[max_row][0].value is None:
            tar_row_num = 1
        else:
            tar_row_num = max_row + 1

    tar_row_num = tar_row_num or tar_sheet.max_row + 1
    if isinstance(row_or_num, tuple):
        src_sheet = row_or_num[0].parent
        row = row_or_num
        num = row[0].row
    else:
        if src_sheet is None:
            raise ValueError('输入行号必须同时输入源表格。')
        row = src_sheet[row_or_num]
        num = row_or_num

    tar_sheet.row_dimensions[tar_row_num].height = src_sheet.row_dimensions[num].height

    for key, cell in enumerate(row, 1):
        col = get_column_letter(key)
        if copy_col_width:
            tar_sheet.column_dimensions[col].width = src_sheet.column_dimensions[col].width
        copy_cell(cell, tar_sheet, tar_row_num, key)


def split_sheet(sheet: Worksheet,
                col: Union[str, int],
                save_path: Union[str, Path],
                begin_row: int = 2,
                keys: Union[str, tuple, list, set] = None) -> None:
    """按照某列内容把xlsx文件拆分成多个文件    \n
    :param sheet: 要拆分的工作表
    :param col: 数据列名或列号
    :param begin_row: 数据行开始行号
    :param save_path: 保存路径
    :param keys: 指定只拆分哪些关键字
    :return: None
    """
    Path(save_path).mkdir(parents=True, exist_ok=True)

    if isinstance(keys, str):
        keys = {keys}

    if keys is not None:
        keys = set(keys)

    if isinstance(col, str):
        col = column_index_from_string(col)

    col -= 1

    if begin_row > 1:
        title_rows = sheet[1:begin_row - 1] if begin_row > 2 else (sheet[1],)
    else:
        begin_row = 1
        title_rows = ()

    rows = list(sheet[begin_row:sheet.max_row])
    rows.sort(key=lambda x: str(x[col].value))

    wb = ws = value = None
    for row in rows:
        now_value = '' if row[col].value is None else str(row[col].value)

        if keys is not None and now_value not in keys:
            continue

        # 如果内容和前一个不同，说明前一个已经采集完，可保存并新建工作簿
        if now_value != value:
            if ws is not None:
                wb.save(f'{save_path}\\{make_valid_name(value)}.xlsx')
                wb.close()

            if keys is not None and ws is not None:
                keys.remove(value)
                if not keys:  # 如果keys内没有内容，说明已经采集完毕，退出循环
                    wb = None
                    break

            wb = Workbook()
            ws = wb.active
            for title in title_rows:
                copy_row(title, ws)

        copy_row(row, ws)
        value = now_value

    if wb is not None:
        wb.save(f'{save_path}\\{make_valid_name(value)}.xlsx')
        wb.close()


def find_in_row(row_or_num: Union[tuple, int],
                key: str,
                sheet: Worksheet = None,
                fuzzy: bool = False,
                return_all: bool = False,
                begin_col: Union[int, str] = 1) -> Union[None, tuple]:
    """在某行中查找一个值，返回单元格坐标或坐标组成的元组             \n
    :param row_or_num: 行元组或行号，输入行号时必须输入sheet
    :param key: 要查找的关键字
    :param sheet: 工作表
    :param fuzzy: 是否模糊查找
    :param return_all: 是否返回全部结果，是则以元组返回多个坐标
    :param begin_col: 开始搜索的列号
    :return: 返回单元格坐标，找不到返回None或空元组
    """
    if isinstance(row_or_num, int):
        if not sheet:
            raise ValueError('用行号定位时必须输入sheet')
        row_or_num = sheet[row_or_num]
    elif not isinstance(row_or_num, tuple):
        raise TypeError('行格式不正确')

    if isinstance(begin_col, str):
        begin_col = column_index_from_string(begin_col)

    return _find_by_fuzzy(row_or_num, key, return_all, begin_col) \
        if fuzzy else _find_by_exact(row_or_num, key, return_all, begin_col)


def find_in_col(col_or_num: Union[tuple, int, str],
                key: str,
                sheet: Worksheet = None,
                fuzzy: bool = True,
                return_all: bool = False,
                begin_row: int = 1) -> Union[None, tuple]:
    """在某列中查找一个值，返回单元格坐标或坐标组成的元组             \n
    :param col_or_num: 列元组或列号
    :param key: 要查找的关键字
    :param sheet: 工作表
    :param fuzzy: 是否模糊查找
    :param return_all: 是否返回全部结果，是则以元组返回多个坐标
    :param begin_row: 开始搜索的行号
    :return: 返回单元格坐标，找不到返回None或空元组
    """
    if isinstance(col_or_num, (str, int)) and not sheet:
        raise ValueError('用行号定位时必须输入sheet')

    if isinstance(col_or_num, int):
        col_or_num = sheet[get_column_letter(col_or_num)]
    elif isinstance(col_or_num, str):
        col_or_num = sheet[col_or_num]
    elif not isinstance(col_or_num, tuple):
        raise TypeError('列格式不正确')

    return _find_by_fuzzy(col_or_num, key, return_all, begin_row) \
        if fuzzy else _find_by_exact(col_or_num, key, return_all, begin_row)


def find_in_sheet(sheet: Worksheet,
                  key: str,
                  fuzzy: bool = True,
                  return_all: bool = False) -> Union[None, tuple]:
    """在工作表中查找一个值，返回单元格坐标或坐标组成的元组             \n
    :param sheet: 工作表
    :param key: 要查找的关键字
    :param fuzzy: 是否模糊查找
    :param return_all: 是否返回全部结果，是则以元组返回多个坐标
    :return: 返回单元格坐标，找不到返回None或空元组
    """
    if return_all:
        locs = []
        for row in sheet.rows:
            loc = _find_by_fuzzy(row, key, True) if fuzzy else _find_by_exact(row, key, True)
            locs.extend(loc)
        return tuple(locs)

    else:
        for row in sheet.rows:
            loc = _find_by_fuzzy(row, key) if fuzzy else _find_by_exact(row, key)
            if loc:
                return loc


def _find_by_fuzzy(row_or_col: tuple,
                   key: str,
                   return_all: bool = False,
                   begin: int = 1) -> Union[tuple, None]:
    """在行或列中以模糊查找的方式查找一个值，返回单元格坐标         \n
    :param row_or_col: 行或列元组
    :param key: 要查找的关键字
    :param return_all: 是否返回全部结果，是则以元组返回多个坐标
    :return: 单元格坐标，找不到返回None
    """
    if begin < 1:
        begin = 1

    if return_all:
        return tuple((i.row, i.column) for i in row_or_col[begin - 1:] if key in str(i.value))
    else:
        for i in row_or_col[begin - 1:]:
            if key in str(i.value):
                return i.row, i.column


def _find_by_exact(row_or_col: tuple,
                   key: str,
                   return_all: bool = False,
                   begin: int = 1) -> Union[tuple, None]:
    """在行或列中以精确查找的方式查找一个值，返回单元格坐标  \n
    :param row_or_col: 行或列元组
    :param key: 要查找的关键字
    :return: 单元格坐标，找不到返回None
    """
    if begin < 1:
        begin = 1

    if return_all:
        return tuple((i.row, i.column) for i in row_or_col[begin - 1:] if key == str(i.value))
    else:
        for i in row_or_col[begin - 1:]:
            if key == str(i.value):
                return i.row, i.column



# -*- coding:utf-8 -*-
# @Author    : g1879
# @date      : 2021/7/10
# @email     : g1879@qq.com
# @File      : files.py
"""
    处理图片的常用方法
    经测试，openCV缩小后图片质量比PIL好
    但openCV写起来比较麻烦
    注意：安装cv2语句是 pip install opencv-python
         安装PIL语句是 pip install pillow
"""
from pathlib import Path
from shutil import copy
from typing import List, Union

from PIL import Image
from cv2 import imencode, imdecode, resize, INTER_AREA
from numpy import fromfile, uint8

from .paths import get_usable_name, get_usable_path


def get_imgs_list(path: str, ignore: Union[str, tuple, list] = None) -> List[Path]:
    """获取所有图片，返回Path对象列表                  \n
    :param path: 文件夹绝对路径
    :param ignore: 忽略的图片格式，如['.jpg']
    :return: 图片绝对路径列表
    """
    if isinstance(ignore, str):
        ignore = (ignore,)

    ignore = ignore or ()
    ignore = set(map(lambda x: x.lower() if x.startswith('.') else f'.{x}'.lower(), ignore))
    formats = {'.jpg', '.png', '.bmp', '.jpeg', '.gif'} - ignore

    return [x for x in Path(path).iterdir() if x.suffix.lower() in formats]


def img_to_jpg(path: str, del_src: bool = True) -> None:
    """将一个图片转为jpg格式                  \n
    :param path: 图片路径
    :param del_src: 是否删除原图片
    :return: None
    """
    img_file = Path(path)

    if img_file.suffix.lower() == '.jpg':
        return

    folder = img_file.parent
    base_name = get_usable_name(str(folder), f'{img_file.stem}.jpg')
    base_name = Path(base_name).stem

    img = Image.open(path)
    img = img.convert('RGB')
    img.save(f'{folder}\\{base_name}.jpg')

    # 删除不是jpg的图片
    if del_src:
        img_file.unlink()


def imgs_to_jpg(path: str, del_src: bool = True) -> None:
    """将传入路径中的所有图片转为jpg                  \n
    :param path: 存放图片的文件夹路径
    :param del_src: 是否删除原图片
    :return: None
    """
    imgs = get_imgs_list(path, ('.jpg',))

    # 执行转换
    for img_path in imgs:
        img_to_jpg(str(img_path), del_src)


def find_min_x(path: Union[Path, str]) -> int:
    """返回路径中最小宽度图片的宽度"""
    return min(tuple(Image.open(x).size[0] for x in get_imgs_list(path)))


def zoom_img(path: Union[Path, str],
             x: int = None,
             y: int = None,
             to_path: str = None,
             rename: str = None) -> Path:
    """按照传入的xy值修改图片大小，如只传入一个值，保持图片比例缩放。可指定保存路径和文件名。  \n
    注意：指定文件名时会覆盖同名文件，否则会自动避免重名。                                 \n
    :param path: 图片路径
    :param x: 压缩后的宽度
    :param y: 压缩后的高度
    :param to_path: 压缩后保存路径，不传入时替换源文件
    :param rename: 保存的文件名。指定文件名时会覆盖同名文件，不指定时会自动重命名。
    :return: 压缩后的路径
    """
    path = Path(path)
    ext = path.suffix
    if to_path:
        Path(to_path).mkdir(parents=True, exist_ok=True)
        new_path = Path(f'{to_path}\\{rename}{ext}') if rename else get_usable_path(f'{to_path}\\{path.name}')
    else:
        new_path = path if not rename else path.parent / f'{rename}{ext}'

    # 读取图片，用这种方式读取才支持中文路径
    img = imdecode(fromfile(path, dtype=uint8), -1)
    old_y, old_x = img.shape[:2]

    # 根据传入的值生成新xy值
    xy = _get_zoom_xy(old_x, old_y, x, y)

    if xy:
        img = resize(img, xy, interpolation=INTER_AREA)
        imencode(ext, img)[1].tofile(new_path)  # 用这种方式保存使中文路径生效
    elif to_path or rename:
        copy(path, new_path)

    return new_path


def crop_img(path: Union[str, Path],
             x: int = None,
             y: int = None,
             to_path: str = None,
             rename: str = None) -> Path:
    """按目标值保持内容比例缩放图片，裁剪超出的部分。如只传入一个值，固定另一个值做裁剪。可指定保存路径和文件名。   \n
    注意：指定文件名时会覆盖同名文件，否则会自动避免重名。                                                  \n
    :param path: 图片路径
    :param x: 裁剪后的宽度
    :param y: 裁剪后的高度
    :param to_path: 裁剪后保存路径，不传入时替换源文件
    :param rename: 保存的文件名。指定文件名时会覆盖同名文件，不指定时会自动重命名。
    :return: 裁剪后的路径
    """
    path = Path(path)
    ext = path.suffix
    if to_path:
        Path(to_path).mkdir(parents=True, exist_ok=True)
        new_path = Path(f'{to_path}\\{rename}{ext}') if rename else get_usable_path(f'{to_path}\\{path.name}')
    else:
        new_path = path if not rename else path.parent / f'{rename}{ext}'

    # 读取图像文件，获取原宽高值
    img = imdecode(fromfile(path, dtype=uint8), -1)
    old_y, old_x = img.shape[:2]
    x = x or old_x
    y = y or old_y

    # 根据比例，缩小图片
    if old_x != x or old_y != y:
        diff = old_x / old_y - x / y
        if diff < 0:
            xy = _get_zoom_xy(old_x, old_y, x=x)
        elif diff > 0:
            xy = _get_zoom_xy(old_x, old_y, y=y)
        elif x > old_x:  # 等比放大图片
            xy = x, y
        else:  # 等比裁剪图片
            xy = False

        # 如果比例有变化，或等比放大，先按长边缩放图片
        if xy:
            old_x, old_y = xy
            img = resize(img, xy, interpolation=INTER_AREA)

        # 根据比例，裁剪图片
        if old_x / old_y <= x / y:
            upper = (old_y - y) // 2
            img = img[upper:upper + y, 0:x]  # 裁剪坐标为[y0:y1, x0:x1]
        else:
            left = (old_x - x) // 2
            img = img[0:y, left:left + x]  # 裁剪坐标为[y0:y1, x0:x1]

        # 保存图片
        imencode(ext, img)[1].tofile(new_path)

    elif to_path or rename:
        copy(path, new_path)

    else:
        new_path = path

    return new_path


def _get_zoom_xy(old_x, old_y, x=None, y=None) -> Union[bool, tuple]:
    """计算缩放图像后的xy值                  \n
    :param old_x: 图像原宽度
    :param old_y: 图像原高度
    :param x: 目标宽度
    :param y: 目标高度
    :return: False或xy值组成的元组
    """
    if (not x and not y) or (x == old_x and y == old_y):
        return False

    if x and not y:
        if old_x == x:
            return False
        else:
            y = int(old_y * x / old_x)
    elif y and not x:
        if old_y == y:
            return False
        else:
            x = int(old_x * y / old_y)

    return x, y



# -*- coding:utf-8 -*-
# win32com安装方法：python -m pip install pypiwin32
from pathlib import Path
from shutil import copy, rmtree
from typing import Union

from win32com import client

from .paths import get_usable_name
from .zip import unzip_7z


def get_pics_in_docx(path: Union[str, Path], to_path: str = None) -> str:
    """获取docx中的图片，存放源路径在独立文件夹中                  \n
    :param path: docx文件路径
    :param to_path: 图片文件夹保存路径，默认和docx文件相同
    :return: 存放图片的文件夹路径
    """
    path = Path(path)

    new_name = get_usable_name(path.parent, path.name)
    zip_path = f'{path.parent}\\{new_name[:new_name.rfind(".")]}.zip'
    copy(str(path), zip_path)

    to_path = to_path or path.parent
    folder_path = f'{to_path}\\{get_usable_name(to_path, path.stem)}'
    unzip_7z(zip_path, folder_path, True, 'overwrite', False)
    Path(zip_path).unlink()

    for i in Path(folder_path).iterdir():
        if i.is_dir():
            rmtree(str(i))
            continue

        if i.suffix not in ('.jpg', '.jpeg', '.png', '.gif', '.bmp'):
            i.unlink()

    return folder_path


def doc_to_docx(path: Union[Path, str], del_src: bool = True) -> bool:
    """把doc格式另存为docx格式                  \n
    :param path: doc文件路径
    :param del_src: 另存之后是否删除源文件
    :return: 是否成功
    """
    word = None

    try:
        word = client.Dispatch("Kwps.Application")  # 调用WPS程序
    except:
        try:
            word = client.gencache.EnsureDispatch('Word.Application')  # 调用Office程序
        except:
            pass

    if word:
        docx = word.Documents.Open(str(path))
        docx.SaveAs(f"{path}x", 51)
        docx.Close()

        if del_src:
            Path(path).unlink()

        return True

    return False



# -*- coding:utf-8 -*-
from time import perf_counter


class Timer(object):
    """用于记录时间间隔的工具"""

    def __init__(self, pin: bool = True, show: bool = True) -> None:
        self.times = []
        if pin:
            self.pin(show)

    def pin(self, show: bool = True) -> None:
        """记录一个时间点
        :param show: 是否打印与上一个时间点的差
        :return: None
        """
        self.times.append(perf_counter())
        if show:
            len_times = len(self.times)
            print(self.times[-1] - self.times[len_times - 2])

    def show(self) -> None:
        """打印所有时间差"""
        for k in range(1, len(self.times)):
            print(f't{k - 1}->t{k}: {self.times[k] - self.times[k - 1]}')



# -*- coding:utf-8 -*-
from pathlib import Path

from cv2 import VideoCapture, imencode, CAP_PROP_POS_MSEC

from .imgs import crop_img


def video_screen(video_path: str,
                 save_path: str = None,
                 img_name: str = None,
                 second: float = 1,
                 x: int = None, y: int = None) -> bool:
    """截取视频某一时间的截图，可调整图片大小                  \n
    :param video_path: 视频的路径
    :param save_path: 图片保存的路径
    :param img_name: 图片文件名，无须后缀，自动保存为jpg
    :param second: 截取第几秒的图像
    :param x: 所保存图片的宽度
    :param y: 所保存图片的高度
    :return:是否截图成功
    """
    if not img_name:
        img_name = Path(video_path).stem
    if not save_path:
        save_path = Path(video_path).parent
    if not img_name.lower().endswith('.jpg'):
        img_name = f'{img_name}.jpg'

    vc = VideoCapture(video_path)  # 读取视频
    vc.set(CAP_PROP_POS_MSEC, second * 1000)  # 设置读取位置
    rval, frame = vc.read()  # 读取当前帧，rval用于判断读取是否成功

    if rval:
        img_path = f'{save_path}\\{img_name}'
        imencode('.jpg', frame)[1].tofile(img_path)

        # 调整图片大小，保持内容比例
        if x or y:
            crop_img(img_path, x, y)

        return True
    else:
        return False



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

# with open("README.md", "r", encoding='utf-8') as fh:
#     long_description = fh.read()

setup(
    name="g1879",
    version="0.4.2",
    author="g1879",
    author_email="g1879@qq.com",
    description="A personal toolkit.",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    license="MIT",
    keywords="g1879",
    url="https://gitee.com/g1879/g1879",
    include_package_data=True,
    packages=find_packages(),
    # install_requires=[
    #     "selenium",
    #     "lxml",
    #     "tldextract",
    #     "requests"
    # ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.8'
)



# -*- coding:utf-8 -*-
from pathlib import Path
from shutil import rmtree
from subprocess import Popen, PIPE
from typing import Union

from .paths import get_usable_name


def unzip_7z(path: Union[str, Path],
             to_path: str,
             no_dir: bool = True,
             file_exists: str = 'skip',
             to_self_dir: bool = True) -> bool:
    """使用7z解压一个压缩文件
    :param path: 压缩文件路径
    :param to_path: 解压目标路径
    :param no_dir: 是否保留文件夹
    :param file_exists: 'skip' 'overwrite' 'rename'
    :param to_self_dir: 是否解压在独立文件夹中
    :return: 是否压缩成功
    """
    if file_exists == 'skip':
        exists = '-aos'
    elif file_exists == 'rename':
        exists = '-aou'
    elif file_exists == 'overwrite':
        exists = '-aoa'
    else:
        raise ValueError("file_exists参数只能传入'skip' 'overwrite' 'rename'")

    mode = 'e' if no_dir else 'x'
    seven_path = Path(__file__).parent / '7z\\7z.exe'
    z = seven_path if seven_path.exists() else '7z'

    if to_self_dir:
        name = Path(path).stem

        if Path(f'{to_path}\\{name}').exists():
            if file_exists == 'skip':
                return False
            elif file_exists == 'rename':
                name = get_usable_name(to_path, name)
            elif file_exists == 'overwrite':
                rmtree(f'{to_path}\\{name}')

        to_path = f'{to_path}\\{name}'

    Path(to_path).mkdir(parents=True, exist_ok=True)

    proc = Popen(f'{z} {mode} {exists} "{path}" "-o{to_path}"',
                 shell=True, stdout=PIPE, stderr=PIPE, bufsize=-1)
    proc.wait()

    return False if proc.communicate()[1] else True


def zip_7z(paths: Union[str, Path, list, tuple],
           to_path: str,
           zip_name: str = None,
           file_exists: str = 'rename') -> Union[bool, str]:
    """打包若干个路径
    :param paths: 要打包的文件路径，可传入多个路径。支持通配符*和?
    :param to_path: 打包文件保存路径
    :param zip_name: 打包文件名称，如果是打包单个路径，可以忽略，不带后缀
    :param file_exists: 'skip' 'overwrite' 'rename' 'merge'
    :return: 是否打包成功
    """
    if isinstance(paths, (str, Path)):
        paths = (paths,)

    if len(set(Path(x).name for x in paths)) < len(paths):
        raise ValueError('路径中有重名文件或文件夹')

    if not zip_name:
        if len(paths) == 1:
            zip_name = f'{Path(paths[0]).stem}.zip'
        else:
            raise ValueError('未传入压缩文件名参数')
    else:
        zip_name = f'{zip_name}.zip'

    full_path = Path(f'{to_path}\\{zip_name}')
    if full_path.exists():
        if file_exists == 'skip':
            return False
        elif file_exists == 'overwrite':
            full_path.unlink()
        elif file_exists == 'rename':
            zip_name = get_usable_name(to_path, zip_name)
        elif file_exists == 'merge':
            pass
        else:
            raise ValueError("只能传入 'skip' 'overwrite' 'rename' 'merge'")

    Path(to_path).mkdir(parents=True, exist_ok=True)
    full_path = f'{to_path}\\{zip_name}'

    files_list = []
    for f in paths:
        file = Path(f)
        if not file.exists():
            raise FileNotFoundError(f'找不到文件：{f}')
        files_list.append(f'"{file}"')
    files_str = ' '.join(files_list)

    seven_path = Path(__file__).parent / '7z\\7z.exe'
    z = seven_path if seven_path.exists() else '7z'

    proc = Popen(f'{z} a "{to_path}\\{zip_name}" {files_str}',
                 shell=True, stdout=PIPE, stderr=PIPE, bufsize=-1)
    proc.wait()

    return False if proc.communicate()[1] else full_path


def zip_a_dir(path: Union[str, Path],
              to_path: str,
              parent_dir: bool = False,
              ignore: Union[Path, str, list, tuple, set] = None,
              file_exists: str = 'rename',
              by_7z: bool = True) -> Union[bool, str]:
    """打包一个文件夹
    :param path: 文件夹路径
    :param to_path: 压缩文件保存位置
    :param parent_dir: 是否保留父级文件夹
    :param ignore: 忽略的文件，可传入一个或多个，如parent_dir为True则此参数无效
    :param file_exists: 'skip' 'overwrite' 'rename' 'merge'
    :param by_7z: 是否使用7z
    :return: 是否打包成功
    """
    path = Path(path)
    if not path.is_dir():
        raise TypeError('传入的不是文件夹')

    zip_name = path.name

    if parent_dir:
        paths = path

    else:
        if ignore is None:
            ignore = tuple()
        elif isinstance(ignore, (str, Path)):
            ignore = (ignore,)

        paths = tuple(file for file in path.iterdir() if file.name not in ignore)

    return zip_7z(paths, to_path, zip_name, file_exists) if by_7z else None

# def unzip(zip_path: str, to_path: str, no_folder=None):
#     乱码文件夹列表 = []  # 记录乱码的文件夹
#     with zipfile.ZipFile(zip_path, 'r') as f:
#         for fn in f.namelist():
#             根目录 = Path(to_path)
#             无乱码路径 = 根目录 / fn.encode("cp437").decode("gbk")
#             解压路径 = Path(f.extract(fn, path=to_path))  # 解压并获取路径对象
#
#             is_file = False if 解压路径.is_dir() else True
#             if not is_file and 无乱码路径 != 解压路径:
#                 乱码文件夹列表.append(解压路径)
#
#             路径 = 无乱码路径.parent / get_valid_name(无乱码路径.parent, 无乱码路径.name) if is_file else 无乱码路径
#             解压路径.rename(路径)  # 重命名乱码文件
#
#             if no_folder and is_file and 无乱码路径.parent != 根目录:
#                 # 移动文件到根目录
#                 文件名 = get_valid_name(根目录, 无乱码路径.name)
#                 无乱码路径.replace(根目录 / 文件名)
#
#     for i in 根目录.iterdir():
#         if i in 乱码文件夹列表 or no_folder:
#             shutil.rmtree(str(i), ignore_errors=True)
#
#
# def do_zip(源路径: str, 重命名: str = None, 包含文件夹: bool = False):
#     目标 = Path(源路径).parent
#     文件夹名 = Path(源路径).name
#     起始 = 0 if 包含文件夹 else len(文件夹名)
#     name = 文件夹名 if not 重命名 else 重命名
#     文件名 = f'{目标 / name}.zip'
#     zipf = zipfile.ZipFile(文件名, 'w')
#     pre_len = len(os.path.dirname(源路径))
#
#     for 父路径, 文件夹s, 文件s in os.walk(源路径):
#         for 文件 in 文件s:
#             文件路径 = os.path.join(父路径, 文件)
#             相对路径 = 文件路径[pre_len:].strip(os.path.sep)  # os.path.sep是路径分隔符，win是\
#             zipf.write(文件路径, 相对路径[起始:])
#
#     zipf.close()
#     return 文件名



# -*- encoding:utf-8 -*-
# @Author    : g1879
# @date      : 2021/6/03
# @email     : g1879@qq.com
# @File      : paths.py
"""
文件、文件夹处理常用方法
"""
from pathlib import Path
from re import sub, search
from shutil import rmtree
from typing import Union, List


def find_path(path: Union[Path, str],
              keys: Union[str, list, tuple, set, dict],
              file_or_dir: str = 'file',
              fuzzy: bool = True,
              find_in_sub: bool = True,
              name_or_suffix: str = 'name',
              match_case: bool = True,
              return_one: bool = True,
              each_key: bool = True,
              deep_first: bool = False) -> Union[Path, List[Path], dict]:
    """根据关键字查找文件或文件夹，返回Path对象或其组成的列表或字典                       \n
    :param path: 在这个文件夹路径下查找
    :param keys: 关键字，可输入多个
    :param file_or_dir: 查找文件还是文件夹，可传入 'file' 'dir' 'both'
    :param fuzzy: 是否模糊匹配
    :param find_in_sub: 是否进入子文件夹查找
    :param name_or_suffix: 在文件名还是后缀名中匹配，可传入 'name' 'suffix' 'full'
    :param match_case: 是否区分大小写
    :param return_one: 只返回第一个结果还是返回全部结果
    :param each_key: 是否分开返回每个key的结果
    :param deep_first: 是否深度优先搜索
    :return: Path对象或其组成的列表或字典
    """
    # -------------- 处理关键字 --------------
    if isinstance(keys, str):
        keys = (keys,)

    keys = map(lambda x: str(x), keys)

    if name_or_suffix == 'suffix':
        keys = map(lambda x: x if x.startswith('.') else f'.{x}', keys)

    if not match_case:
        keys = map(lambda x: x.lower(), keys)

    keys = set(keys)

    # -------------- 执行查找操作 --------------
    results = {x: None for x in keys} if each_key else []

    folders = []
    for item in Path(path).iterdir():
        # 判断模式是否和当前类型一致，不一致则跳过该文件（夹）
        if ((file_or_dir == 'both')
                or (file_or_dir == 'file' and item.is_file())
                or (file_or_dir == 'dir' and item.is_dir())):
            # 准备待匹配的部分
            if name_or_suffix == 'name':
                find_str = item.stem
            elif name_or_suffix == 'suffix':
                find_str = item.suffix
            elif name_or_suffix == 'full':
                find_str = item.name
            else:
                raise ValueError("name_or_suffix参数只能传入'name', 'suffix', 'full'")

            # 若不区分大小写，全部转为小写
            if not match_case:
                find_str = find_str.lower()

            # 对关键字进行匹配
            is_it = None
            item_keys = []
            for key in tuple(keys):
                if (fuzzy and key in find_str) or (not fuzzy and find_str == key):
                    is_it = True
                    item_keys.append(key)

                    if return_one:
                        keys.remove(key)

                    if not each_key:
                        break

            # 若匹配成功，即该文件（夹）为结果（之一）
            if is_it:
                # 用字典返回每个key的结果
                if each_key:
                    for k in item_keys:
                        # 如果只要第一个结果，且该key值已经有结果，就跳到下一个
                        if return_one and results.get(k, None):
                            continue

                        if results.get(k, None) is None:
                            results[k] = item if return_one else [item]
                        else:
                            results[k].append(item)

                # 用列表或Path对象返回全部结果
                else:
                    if return_one:
                        return item
                    else:
                        results.append(item)

        # -------------- 如果是文件夹，调用自己，进入下层 --------------
        if item.is_dir() and find_in_sub and keys:
            # 广度优先，记录文件夹，等下再处理
            if not deep_first:
                folders.append(item)

            # 深度优先，遇到文件夹即时进入查找
            else:
                sub_results = find_path(item, keys, file_or_dir, fuzzy, find_in_sub, name_or_suffix,
                                        match_case, return_one, each_key, deep_first)
                results = _merge_results(results, sub_results, each_key, return_one)

    # 广度优先，上面判断完文件，再逐个进入该层文件夹查找
    if not deep_first:
        for item in folders:
            sub_results = find_path(item, keys, file_or_dir, fuzzy, find_in_sub, name_or_suffix,
                                    match_case, return_one, each_key, deep_first)
            results = _merge_results(results, sub_results, each_key, return_one)

    return results


def clean_dir(path: str, ignore: Union[str, list, tuple] = None) -> None:
    """清空一个文件夹，除了ignore里的文件和文件夹                  \n
    :param path: 要清空的文件夹路径
    :param ignore: 忽略列表或文件（夹）全名
    :return: None
    """
    if isinstance(ignore, str):
        ignore = (ignore,)

    for f in Path(path).iterdir():
        if not ignore or f.name not in ignore:
            if f.is_file():
                f.unlink()
            elif f.is_dir():
                rmtree(f, True)


def get_usable_path(path: Union[str, Path]) -> Path:
    """检查文件或文件夹是否有重名，并返回可以使用的路径           \n
    :param path: 文件或文件夹路径
    :return: 可用的路径，Path对象
    """
    path = Path(path)
    parent = path.parent
    path = parent / make_valid_name(path.name)
    name = path.stem if path.is_file() else path.name
    ext = path.suffix if path.is_file() else ''

    first_time = True

    while path.exists():
        r = search(r'(.*)_(\d+)$', name)

        if not r or (r and first_time):
            src_name, num = name, '1'
        else:
            src_name, num = r.group(1), int(r.group(2)) + 1

        name = f'{src_name}_{num}'
        path = parent / f'{name}{ext}'
        first_time = None

    return path


def get_usable_name(path: Union[str, Path], name: str) -> str:
    """检查文件或文件夹是否重名，并返回可以使用的名称                  \n
    :param path: 文件夹路径
    :param name: 要检查的名称
    :return: 可用的文件名
    """
    return get_usable_path(Path(path) / name).name


def get_desktop() -> str:
    """获取桌面路径"""
    from winreg import QueryValueEx, OpenKey, HKEY_CURRENT_USER
    return QueryValueEx(
        OpenKey(HKEY_CURRENT_USER, r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'),
        "Desktop")[0]


def get_size(path: Union[str, Path], unit: str = None) -> Union[int, float]:
    """获取文件或文件夹大小                  \n
    :param path: 文件或文件夹路径
    :param unit: 单位，'k' 'm' 'g'
    :return: 文件或文件夹大小，非占用空间
    """
    from os import walk
    from os.path import getsize, join, isdir, exists

    path = str(path)

    if not exists(path):
        raise FileNotFoundError
    elif isdir(path):
        size = 0
        for root, dirs, files in walk(path):
            size += sum([getsize(join(root, name)) for name in files])
    else:
        size = getsize(path)

    unit = unit.lower() if unit else None
    if unit == 'k':
        return size / 1024
    elif unit == 'm':
        return size / 1048576
    elif unit == 'g':
        return size / 1073741824
    else:
        return size


def get_long(txt) -> int:
    """返回字符串中字符个数（一个汉字是2个字符）          \n
    :param txt: 字符串
    :return: 字符个数
    """
    txt_len = len(txt)
    return int((len(txt.encode('utf-8')) - txt_len) / 2 + txt_len)


def make_valid_name(full_name: str) -> str:
    """获取有效的文件名                  \n
    :param full_name: 文件名
    :return: 可用的文件名
    """
    # ----------------去除前后空格----------------
    full_name = full_name.strip()

    # ----------------使总长度不大于255个字符（一个汉字是2个字符）----------------
    r = search(r'(.*)(\.[^.]+$)', full_name)  # 拆分文件名和后缀名
    if r:
        name, ext = r.group(1), r.group(2)
        ext_long = len(ext)
    else:
        name, ext = full_name, ''
        ext_long = 0

    while get_long(name) > 255 - ext_long:
        name = name[:-1]

    full_name = f'{name}{ext}'

    # ----------------去除不允许存在的字符----------------
    return sub(r'[<>/\\|:*?\n]', ' ', full_name)


def _merge_results(par_results: Union[dict, list],
                   sub_results: Union[dict, list, Path],
                   each_key: bool,
                   return_one: bool) -> Union[dict, list, Path]:
    """合并查找结果                                      \n
    :param par_results: 父级查找结果
    :param sub_results: 子文件夹查找结果
    :param each_key: 是否分开返回每个key的结果
    :param return_one: 只返回第一个结果还是返回全部结果
    :return: 合并后的结果
    """
    # 用字典返回每个关键字结果
    if each_key and any(sub_results.values()):
        # 将父字典和子字典的结果合并
        for k in tuple(x for x in sub_results if sub_results[x]):
            if par_results.get(k, None) is None:
                par_results[k] = sub_results[k] if return_one else [sub_results[k]]
            else:
                par_results[k].extend(sub_results[k])

    # 用列表或Path对象返回查找结果
    elif not each_key and sub_results:
        if return_one:
            return sub_results
        else:
            par_results.extend(sub_results)

    return par_results
